#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <errno.h>
#include <ctype.h>

#include "ftp.h"

#define FTP_WANT_FLAGS_OR_POLL(var, flags, poll) \
	do { \
	if ((var & flags) == 0) { \
		poll = flags; \
		return FTP_POLL; \
	} \
	} while (0)

enum {
	FTP_LINE_INVALID = 0,
	FTP_LINE_CODE,
	FTP_LINE_CONTINUE
};

enum {
	FTP_INTERNAL_FREE = 0,
	FTP_INTERNAL_WRITE,
	FTP_INTERNAL_READ,
	
	FTP_INTERNAL_DATA_WRITE,
	FTP_INTERNAL_DATA_READ
};

struct _FTPSession {
	/* El socket de control */
	int csock;
	
	/* El socket de datos, si está abierto */
	int dsock;
	
	/* Estado de la sessión en general */
	int estado;
	
	/* Estado interno, lectura, escritura, sobre control o datos */
	int internal;
	
	/* Controla si cuando enviamos un comando debemos vigilar el canal de datos */
	int command_data_dir;
	
	/* La conexión con el servidor */
	struct sockaddr_storage server;
	
	/* El buffer de lectura de la red */
	char read_buffer[8192];
	size_t read_count;
	int ceof;
	
	/* Variable que controla el poller */
	int poll;
	
	/* El último código de 3 digitos del último comando enviado */
	int last_code;
	
	/* Utilizado para saber si sigo buscando lineas de continuación de un comando multilineas */
	char cont_code[6];
	char *user;
	char *passwd;
	char last_cmd[8192];
	char last_message[1024];
	
	/* El stream que usaremos para enviar o recibir datos */
	FILE *data_stream;
	
	/* Función para llamar cuando lea entre lineas */
	FTPLineFunc line_func;
	void *line_func_data;
};

int ftp_session_read_bytes (FTPSession *session);
int ftp_session_readline (FTPSession *session, char *buffer, int max);

int _ftp_session_internal_connect_cont (FTPSession *session);
int _ftp_session_internal_banner_cont (FTPSession *session);
int _ftp_session_internal_send_user (FTPSession *session);
int _ftp_session_internal_user_code (FTPSession *session);
int _ftp_session_internal_send_pass (FTPSession *session);
int _ftp_session_internal_pass_code (FTPSession *session);
int _ftp_session_internal_send_cmd (FTPSession *session);
int _ftp_session_internal_cmd_code (FTPSession *session);
int _ftp_session_internal_send_close (FTPSession *session);
int _ftp_session_internal_close_code (FTPSession *session);
int _ftp_session_internal_send_pasv (FTPSession *session);
int _ftp_session_internal_pasv_code (FTPSession *session);
int _ftp_session_internal_data_connect_cont (FTPSession *session);
int _ftp_session_internal_cmd_data_read (FTPSession *session);
int _ftp_session_internal_cmd_data_write (FTPSession *session);

FTPSession *ftp_session_create (void) {
	FTPSession *ftp;
	
	ftp = (FTPSession *) malloc (sizeof (FTPSession));
	
	if (ftp == NULL) return NULL;
	
	memset (ftp, 0, sizeof (FTPSession));
	
	ftp->csock = -1;
	ftp->dsock = -1;
	
	return ftp;
}

void ftp_session_free (FTPSession *session) {
	ftp_session_close_and_clean (session);
	
	free (session);
}

void ftp_session_close_and_clean (FTPSession *session) {
	if (session->user != NULL) {
		free (session->user);
		session->user = NULL;
	}
	
	if (session->passwd != NULL) {
		free (session->passwd);
		session->passwd = NULL;
	}
	
	if (session->csock >= 0) {
		close (session->csock);
		session->csock = -1;
	}
	
	if (session->dsock >= 0) {
		close (session->dsock);
		session->dsock = -1;
	}
	
	session->internal = FTP_INTERNAL_FREE;
	session->estado = FTP_STATE_CLOSED;
}

static int _ftp_session_validate_response_line (FTPSession *session, char *buffer, int size) {
	if (size < 4 && session->cont_code[0] == 0) {
		return FTP_LINE_INVALID;
	}
	
	if (session->cont_code[0] != 0) {
		/* Estamos leyendo una continuación, si son digitos más espacio, es la linea final */
		if (size >= 4 && strncmp (session->cont_code, buffer, 3) == 0 && buffer[3] == ' ') {
			session->cont_code[0] = 0; /* Ya no es continuación */
			return FTP_LINE_CODE;
		} else {
			/* Lo que sea la linea, es continuación de lo anterior */
			
			/* Si el usuario instaló una función de leer entre lineas, llamarla */
			if (session->line_func != NULL) {
				session->line_func (session, buffer, session->line_func_data);
			}
			return FTP_LINE_CONTINUE;
		}
	} else {
		if (!isdigit (buffer[0]) || !isdigit (buffer[1]) || !isdigit (buffer[2])) {
			return FTP_LINE_INVALID;
		}
		
		if (buffer[3] == '-') {
			/* Empieza una continuación multi-linea */
			session->cont_code[0] = buffer[0];
			session->cont_code[1] = buffer[1];
			session->cont_code[2] = buffer[2];
			session->cont_code[3] = 0;
			
			/* Si el usuario instaló una función de leer entre lineas, llamarla */
			if (session->line_func != NULL) {
				session->line_func (session, buffer, session->line_func_data);
			}
			
			return FTP_LINE_CONTINUE;
		} else {
			return FTP_LINE_CODE;
		}
	}
}

int ftp_session_connect (FTPSession *session, struct sockaddr *server, socklen_t addr_len) {
	int c;
	int res;
	
	session->poll = FTP_WANT_NONE;
	
	c = socket (server->sa_family, SOCK_STREAM, 0);
	
	if (c < 0) {
		return FTP_ERROR;
	}
	
	fcntl (c, F_SETFL, O_NONBLOCK);
	
	session->estado = FTP_STATE_TRYING_CONNECT;
	memcpy (&session->server, server, addr_len);
	
	res = connect (c, server, addr_len);
	
	if (res < 0) {
		if (errno == EINPROGRESS) {
			session->csock = c;
			session->poll = FTP_WANT_POLL_OUT;
			session->internal = FTP_INTERNAL_WRITE;
			session->estado = FTP_STATE_TRYING_CONNECT;
			return FTP_POLL;
		} else {
			session->csock = -1;
			close (c);
			
			return FTP_ERROR;
		}
	} else {
		/* Conexión exitosa */
		session->csock = c;
		session->internal = FTP_INTERNAL_READ;
		session->estado = FTP_STATE_TRYING_CONNECT;
		session->read_count = 0;
		return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
	}
}

int ftp_session_async_continue (FTPSession *session, int what) {
	char linea[8192];
	size_t linea_size;
	int res;
	
	if (session->internal == FTP_INTERNAL_FREE) {
		return FTP_ERROR;
	}
	
	if (session->internal == FTP_INTERNAL_DATA_READ) {
		FTP_WANT_FLAGS_OR_POLL (what, FTP_WANT_POLL_DATA_IN, session->poll);
		
		switch (session->estado) {
			case FTP_STATE_CMD:
				return _ftp_session_internal_cmd_data_read (session);
		}
	} else if (session->internal == FTP_INTERNAL_DATA_WRITE) {
		FTP_WANT_FLAGS_OR_POLL (what, FTP_WANT_POLL_DATA_OUT, session->poll);
		
		switch (session->estado) {
			case FTP_STATE_PASV:
				return _ftp_session_internal_data_connect_cont (session);
			case FTP_STATE_CMD:
				return _ftp_session_internal_cmd_data_write (session);
		}
	} else if (session->internal == FTP_INTERNAL_WRITE) {
		FTP_WANT_FLAGS_OR_POLL (what, FTP_WANT_POLL_OUT, session->poll);
		
		/* Se logró el evento esperado */
		switch (session->estado) {
			case FTP_STATE_TRYING_CONNECT:
				return _ftp_session_internal_connect_cont (session);
			case FTP_STATE_LOGIN_USER:
				return _ftp_session_internal_send_user (session);
			case FTP_STATE_LOGIN_PASS:
				return _ftp_session_internal_send_pass (session);
			case FTP_STATE_CMD:
				return _ftp_session_internal_send_cmd (session);
			case FTP_STATE_CLOSING:
				return _ftp_session_internal_send_close (session);
			case FTP_STATE_PASV:
				return _ftp_session_internal_send_pasv (session);
		}
	} else if (session->internal == FTP_INTERNAL_READ) {
		FTP_WANT_FLAGS_OR_POLL (what, FTP_WANT_POLL_IN, session->poll);
		
		/* Cuando sea una lectura, esperamos una linea con un código de respuesta, el que sea */

ftp_continue_read_line_again:
		/* Leer una linea */
		linea_size = ftp_session_readline (session, linea, sizeof (linea));
		if (linea_size == FTP_POLL) {
			/* Si al intentar leer, no hay linea, revisar si podemos leer algunos bytes */
			res = ftp_session_read_bytes (session);
			if (res == FTP_ERROR) {
				ftp_session_close_and_clean (session);
				return FTP_ERROR;
			} else if (res == FTP_POLL) {
				/* No hay bytes, hacer polling */
				session->poll = FTP_WANT_POLL_IN;
				return FTP_POLL;
			} else if (session->ceof == 1) {
				/* Cierre de la conexión, limpiar y cerrar */
				if (session->estado == FTP_STATE_CLOSING) {
					ftp_session_close_and_clean (session);
					return FTP_SUCCESS;
				} else {
					ftp_session_close_and_clean (session);
					return FTP_ERROR;
				}
			}
			/* Con los bytes restantes de la lectura previa, intentar leer la linea */
			linea_size = ftp_session_readline (session, linea, sizeof (linea));
		}
		
		/* Procesar la linea */
		res = _ftp_session_validate_response_line (session, linea, linea_size);
		
		if (res == FTP_LINE_INVALID) {
			/* Esperaba una respuesta numérica */
			ftp_session_close_and_clean (session);
			return FTP_ERROR;
		} else if (res == FTP_LINE_CONTINUE) {
			/* Llamarnos recursivamente para procesar la siguiente linea */
			goto ftp_continue_read_line_again;
		}
		/* Linea buena, guardar el estado y el mensaje de la linea */
		/* Estamos conectados */
		strncpy (session->last_message, &linea[4], sizeof (session->last_message) - 1);
		session->last_code = ((linea[0] - '0') * 100) + ((linea[1] - '0') * 10) + (linea[2] - '0');
		/* Correr la función correcta para pasar al siguiente estado */
		switch (session->estado) {
			case FTP_STATE_TRYING_CONNECT:
				return _ftp_session_internal_banner_cont (session);
			case FTP_STATE_LOGIN_USER:
				return _ftp_session_internal_user_code (session);
			case FTP_STATE_LOGIN_PASS:
				return _ftp_session_internal_pass_code (session);
			case FTP_STATE_CMD:
				return _ftp_session_internal_cmd_code (session);
			case FTP_STATE_CLOSING:
				return _ftp_session_internal_close_code (session);
			case FTP_STATE_PASV:
				return _ftp_session_internal_pasv_code (session);
		}
	}
}

int _ftp_session_internal_connect_cont (FTPSession *session) {
	int optval;
	socklen_t optlen;
	int res;
	
	optval = -1;
	
	optlen = sizeof (optval);
	
	res = getsockopt (session->csock, SOL_SOCKET, SO_ERROR, &optval, &optlen);

	if (res < 0) {
		ftp_session_close_and_clean (session);

		return FTP_ERROR;
	}
	
	/* FIXME: Revisar el valor de optval */
	
	/* Pasamos a la lectura del banner */
	session->internal = FTP_INTERNAL_READ;
	session->estado = FTP_STATE_TRYING_CONNECT;
	session->read_count = 0;
	return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
}

int _ftp_session_internal_banner_cont (FTPSession *session) {
	if (session->last_code >= 200 && session->last_code < 300) {
		session->internal = FTP_INTERNAL_FREE;
		session->estado = FTP_STATE_CONNECTED;
		return FTP_SUCCESS;
	} else {
		ftp_session_close_and_clean (session);
		return FTP_ERROR;
	}
}

int _ftp_session_internal_send_user (FTPSession *session) {
	char linea[1024];
	int res, linea_size;
	
	linea_size = snprintf (linea, sizeof (linea), "USER %s\r\n", session->user);
	res = write (session->csock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = FTP_WANT_POLL_OUT;
		session->internal = FTP_INTERNAL_WRITE;
		return FTP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		ftp_session_close_and_clean (session);
		
		return FTP_ERROR;
	}
	/* Envio completo */
	session->internal = FTP_INTERNAL_READ;
	
	return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
}

int _ftp_session_internal_user_code (FTPSession *session) {
	if (session->last_code >= 300 && session->last_code < 400) {
		session->estado = FTP_STATE_LOGIN_PASS;
		session->internal = FTP_INTERNAL_WRITE;
		return ftp_session_async_continue (session, FTP_WANT_POLL_OUT);
	} else if (session->last_code >= 200 && session->last_code < 300) {
		session->internal = FTP_INTERNAL_FREE;
		session->estado = FTP_STATE_LOGGED;
		
		if (session->passwd) {
			free (session->passwd);
			session->passwd = NULL;
		}
		return FTP_SUCCESS;
	} else if (session->last_code >= 530 && session->last_code < 540) {
		session->estado = FTP_STATE_CONNECTED;
		session->internal = FTP_INTERNAL_FREE;
		return FTP_SUCCESS;
	} else {
		ftp_session_close_and_clean (session);
		return FTP_ERROR;
	}
}

int _ftp_session_internal_send_pass (FTPSession *session) {
	char linea[1024];
	int res, linea_size;
	
	linea_size = snprintf (linea, sizeof (linea), "PASS %s\r\n", session->passwd);
	res = write (session->csock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = FTP_WANT_POLL_OUT;
		session->internal = FTP_INTERNAL_WRITE;
		return FTP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		ftp_session_close_and_clean (session);
		
		return FTP_ERROR;
	}
	/* Envio completo */
	session->internal = FTP_INTERNAL_READ;
	
	return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
}

int _ftp_session_internal_pass_code (FTPSession *session) {
	if (session->last_code >= 200 && session->last_code < 300) {
		session->estado = FTP_STATE_LOGGED;
		session->internal = FTP_INTERNAL_FREE;
	} else {
		session->estado = FTP_STATE_CONNECTED;
		session->internal = FTP_INTERNAL_FREE;
	}
	return FTP_SUCCESS;
}

int _ftp_session_internal_send_cmd (FTPSession *session) {
	int res, linea_size;
	
	linea_size = strlen (session->last_cmd);
	res = write (session->csock, session->last_cmd, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = FTP_WANT_POLL_OUT;
		session->internal = FTP_INTERNAL_WRITE;
		return FTP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		ftp_session_close_and_clean (session);
		
		return FTP_ERROR;
	}
	/* Envio completo */
	session->internal = FTP_INTERNAL_READ;
	
	return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
}

int _ftp_session_internal_cmd_code (FTPSession *session) {
	if (session->command_data_dir != FTP_DATA_NONE && session->dsock != -1) {
		if (session->last_code >= 100 && session->last_code < 200) {
			/* Como pidieron leer/escribir datos en el canal de datos, pasar a internal_data read/write */
			if (session->command_data_dir == FTP_DATA_READ) {
				session->internal = FTP_INTERNAL_DATA_READ;
				
				return ftp_session_async_continue (session, FTP_WANT_POLL_DATA_IN);
			} else {
				session->internal = FTP_INTERNAL_DATA_WRITE;
				
				return ftp_session_async_continue (session, FTP_WANT_POLL_DATA_OUT);
			}
		} else {
			/* Error es error */
			session->estado = FTP_STATE_LOGGED;
			session->internal = FTP_INTERNAL_FREE;
		}
	} else if (session->last_code >= 100 && session->last_code < 200) {
		/* Si no pidieron leer datos, entonces, un código 100 significa esperar por un 200 */
		session->internal = FTP_INTERNAL_READ;
		
		return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
	} else {
		session->estado = FTP_STATE_LOGGED;
		session->internal = FTP_INTERNAL_FREE;
	}
	
	return FTP_SUCCESS;
}

int _ftp_session_internal_cmd_data_read (FTPSession *session) {
	/* Leer datos de socket de datos y enviar al flujo que nos pasaron */
	int res;
	char buffer[8192];
	/* Tratar de sacar tantos datos como nos sea posible hasta golpear un EOF o error */
	do {
		res = recv (session->dsock, buffer, sizeof (buffer), 0);
		
		if (res == 0) {
			/* Fin de archivo sobre el socket de datos */
			close (session->dsock);
			session->dsock = -1;
			session->data_stream = NULL;
			
			/* Pasar a leer el 200 del comando */
			session->internal = FTP_INTERNAL_READ;
			
			return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
		} else if (res < 0) {
			if (errno == EAGAIN || errno == EINTR) {
				/* Se acabaron los datos por leer, continuar */
				session->poll = FTP_WANT_POLL_DATA_IN;
				return FTP_POLL;
			}
			
			close (session->dsock);
			session->dsock = -1;
			session->data_stream = NULL;
			
			/* ¿error en el canal de datos? ¿Qué hago? */
			session->internal = FTP_INTERNAL_READ;
			
			return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
		}
		
		if (session->data_stream != NULL) {
			fwrite (buffer, 1, res, session->data_stream);
		}
	} while (1);
}

int _ftp_session_internal_cmd_data_write (FTPSession *session) {
	int res;
	char buffer[8192];
	size_t consumed;
	long pos;
	
	if (session->data_stream == NULL) {
		/* Si no hay flujo, nada que mandar. Cerrar canal de datos */
		close (session->dsock);
		session->dsock = -1;
		
		/* Pasar a leer el 200 del comando */
		session->internal = FTP_INTERNAL_READ;
		
		return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
	}

ftp_cmd_data_write_again:
	/* Primero hacer una lectura del flujo */
	pos = ftell (session->data_stream);
	consumed = fread (buffer, 1, sizeof (buffer), session->data_stream);
	if (consumed == 0) {
		//if (feof (session->data_stream) || ferror (session->data_stream)) {
			/* Fin de archivo, continuar con los comandos */
			close (session->dsock);
			session->dsock = -1;
			session->data_stream = NULL;
			
			/* Pasar a leer el 200 del comando */
			session->internal = FTP_INTERNAL_READ;
			
			return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
		//}
	}
	
	/* Tratar de escribir la cantidad leida */
	res = send (session->dsock, buffer, consumed, 0);
	if (res > 0) {
		/* Reajustar el flujo de entrada */
		pos = pos + res;
		
		fseek (session->data_stream, pos, SEEK_SET);
		
		/* Reintentar enviar datos */
		goto ftp_cmd_data_write_again;
	} else if (res <= 0) {
		if (errno == EAGAIN || errno == EINTR) {
			/* Bloquearía, reintentar después */
			fseek (session->data_stream, pos, SEEK_SET);
			
			session->poll = FTP_WANT_POLL_DATA_OUT;
			return FTP_POLL;
		}
		
		/* En otro error, se cerró la conexión de datos */
		close (session->dsock);
		session->dsock = -1;
		session->data_stream = NULL;
		
		session->internal = FTP_INTERNAL_READ;
		
		return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
	}
}

int _ftp_session_internal_send_close (FTPSession *session) {
	char *linea;
	int res, linea_size;
	
	linea = "QUIT\r\n";
	linea_size = strlen (linea);
	res = write (session->csock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = FTP_WANT_POLL_OUT;
		session->internal = FTP_INTERNAL_WRITE;
		return FTP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		ftp_session_close_and_clean (session);
		
		return FTP_ERROR;
	}
	/* Envio completo */
	session->internal = FTP_INTERNAL_READ;
	
	return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
}

int _ftp_session_internal_close_code (FTPSession *session) {
	if (session->last_code >= 200 && session->last_code < 300) {
		session->internal = FTP_INTERNAL_FREE;
		
		/* Intencionalmente no cambiamos de estado para que nos llegue el cierre */
		return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
	} else {
		ftp_session_close_and_clean (session);
	}
	
	return FTP_SUCCESS;
}

int _ftp_session_internal_send_pasv (FTPSession *session) {
	char *linea;
	int res, linea_size;
	
	/* Elegir el tipo correcto de EPSV o PASV basado en el tipo de socket */
	if (session->server.ss_family == AF_INET) {
		linea = "PASV\r\n";
	} else {
		linea = "EPSV\r\n";
	}
	linea_size = strlen (linea);
	res = write (session->csock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = FTP_WANT_POLL_OUT;
		session->internal = FTP_INTERNAL_WRITE;
		return FTP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		ftp_session_close_and_clean (session);
		
		return FTP_ERROR;
	}
	/* Envio completo */
	session->internal = FTP_INTERNAL_READ;
	
	return ftp_session_async_continue (session, FTP_WANT_POLL_IN);
}

static socklen_t _ftp_parse_port_pasv (int family, char *linea, struct sockaddr *dest) {
	char *str;
	unsigned char v[6];
	long n;
	struct sockaddr_in *ipv4;
	struct sockaddr_in6 *ipv6;
	
	str = strchr (linea, '(');
	
	if (str == NULL) {
		return 0;
	}
	
	str++;
	
	if (family == AF_INET) {
		ipv4 = (struct sockaddr_in *) dest;
		ipv4->sin_family = AF_INET;
		n = sscanf (str, "%hhu,%hhu,%hhu,%hhu,%hhu,%hhu", &v[2], &v[3], &v[4], &v[5], &v[0], &v[1]);
		
		if (n < 6) {
			return 0;
		}
		if (v[2] != 0 || v[3] != 0 || v[4] != 0 || v[5] != 0) {
			ipv4->sin_addr.s_addr = htonl ((v[2] << 24) | (v[3] << 16) | (v[4] << 8) | v[5]);
		}
		ipv4->sin_port = htons ((v[0] << 8) | v[1]);
		
		return sizeof (struct sockaddr_in);
	} else if (family == AF_INET6) {
		ipv6 = (struct sockaddr_in6 *) dest;
		ipv6->sin6_family = AF_INET6;
		
		v[0] = sscanf (str, "|||%ld", &n);
		
		if (v[0] < 1) {
			return 0;
		}
		
		ipv6->sin6_port = htons (n);
		
		return sizeof (struct sockaddr_in6);
	}
	
	return 0;
}

int _ftp_session_internal_pasv_code (FTPSession *session) {
	int c;
	struct sockaddr_storage pasv_addr;
	socklen_t len;
	int res;
	
	if (session->last_code >= 200 && session->last_code < 300) {
		if (session->server.ss_family == AF_INET) {
			memcpy (&pasv_addr, &session->server, sizeof (struct sockaddr_in));
		} else if (session->server.ss_family == AF_INET6) {
			memcpy (&pasv_addr, &session->server, sizeof (struct sockaddr_in6));
		}
		
		len = _ftp_parse_port_pasv (session->server.ss_family, session->last_message, (struct sockaddr *) &pasv_addr);
		
		if (len == 0) {
			/* Falló el pasivo, regresar a logueado */
			session->internal = FTP_INTERNAL_FREE;
			session->estado = FTP_STATE_LOGGED;
			
			return FTP_ERROR;
		}
		
		/* Parsear el puerto desde el mensaje recibido */
		c = socket (pasv_addr.ss_family, SOCK_STREAM, 0);
		if (c < 0) {
			session->internal = FTP_INTERNAL_FREE;
			session->estado = FTP_STATE_LOGGED;
			return FTP_ERROR;
		}
		
		fcntl (c, F_SETFL, O_NONBLOCK);
		
		res = connect (c, (struct sockaddr *) &pasv_addr, len);
		
		if (res < 0) {
			if (errno == EINPROGRESS) {
				session->dsock = c;
				session->poll = FTP_WANT_POLL_DATA_OUT;
				session->internal = FTP_INTERNAL_DATA_WRITE;
				return FTP_POLL;
			} else {
				session->dsock = -1;
				close (c);
				
				session->internal = FTP_INTERNAL_FREE;
				session->estado = FTP_STATE_LOGGED;
				return FTP_ERROR;
			}
		} else {
			/* Conexión exitosa */
			session->dsock = c;
			session->internal = FTP_INTERNAL_FREE;
			session->estado = FTP_STATE_LOGGED;
			session->read_count = 0;
			return FTP_SUCCESS;
		}
	} else {
		/* En el caso de un pasivo con un código de error, simplemente regresar al estado original */
		session->internal = FTP_INTERNAL_FREE;
		session->estado = FTP_STATE_LOGGED;
		
		return FTP_ERROR;
	}
	
	return FTP_SUCCESS;
}

int _ftp_session_internal_data_connect_cont (FTPSession *session) {
	int optval;
	socklen_t optlen;
	int res;
	
	optval = -1;
	
	optlen = sizeof (optval);

	res = getsockopt (session->dsock, SOL_SOCKET, SO_ERROR, &optval, &optlen);

	if (res < 0) {
		close (session->dsock);
		session->dsock = -1;

		return FTP_ERROR;
	}
	
	if (optval != 0) {
		close (session->dsock);
		session->dsock = -1;

		return FTP_ERROR;
	}
	
	/* Tenemos socket para data */
	session->internal = FTP_INTERNAL_FREE;
	session->estado = FTP_STATE_LOGGED;
	return FTP_SUCCESS;
}

int ftp_session_read_bytes (FTPSession *session) {
	int res;
	if (session->read_count >= sizeof (session->read_buffer)) {
		return FTP_ERROR;
	}
	
	/* Leer tantos bytes como tenga el buffer */
	res = read (session->csock, &session->read_buffer[session->read_count], sizeof (session->read_buffer) - session->read_count);
	
	if (res == 0) {
		session->ceof = 1;
	} else if (res < 0) {
		if (errno == EINTR || errno == EAGAIN) {
			return FTP_POLL;
		}
		return FTP_ERROR;
	} else {
		session->read_count += res;
	}
	
	return FTP_SUCCESS;
}

int ftp_session_readline (FTPSession *session, char *buffer, int max) {
	char *end;
	int line_size = 0;
	
	if (session->read_count > 0) {
		/* Ya había otros bytes en el buffer, procesarlos para buscar el salto de linea aquí */
		line_size = (max >= session->read_count) ? session->read_count : max - 1;
		
		end = memccpy (buffer, session->read_buffer, '\n', line_size);
		
		if (end != NULL) {
			line_size = end - buffer;
		}
		
		buffer[line_size] = 0;
		
		/* Consumir x bytes de la entrada */
		if (line_size < session->read_count) {
			memmove (session->read_buffer, &session->read_buffer[line_size], session->read_count - line_size);
			
			session->read_count -= line_size;
		} else {
			session->read_buffer[0] = 0;
			session->read_count = 0;
		}
		
		if (end != NULL) {
			
			if (buffer[line_size - 2] == '\r') {
				/* Quitar el \r */
				buffer[line_size - 2] = '\n';
				buffer[line_size - 1] = 0;
				line_size--;
			}
			
			return line_size;
		}
	}
	
	return FTP_POLL;	
}

int ftp_session_get_state (FTPSession *session) {
	return session->estado;
}

/* Las funciones públicas */
int ftp_session_login (FTPSession *session, const char *user, const char *passwd) {
	if (session->estado != FTP_STATE_CONNECTED || session->internal != FTP_INTERNAL_FREE) {
		return FTP_ERROR;
	}
	
	/* Copiar el usuario y contraseña */
	session->user = strdup (user);
	if (passwd != NULL) {
		session->passwd = strdup (passwd);
	} else {
		session->passwd = NULL;
	}
	
	session->estado = FTP_STATE_LOGIN_USER;
	session->internal = FTP_INTERNAL_WRITE;
	return ftp_session_async_continue (session, FTP_WANT_POLL_OUT);
}

int ftp_session_send_command (FTPSession *session, char *command, int data_dir) {
	int len;
	
	if (session->estado != FTP_STATE_LOGGED || session->internal != FTP_INTERNAL_FREE) {
		return FTP_ERROR;
	}
	
	/* Sanitizar el comando y asegurarnos que contenga el \r\n */
	len = strlen (command);
	if (command[len - 2] == '\r' && command[len - 1] == '\n') {
		if (len + 1 > sizeof (session->last_cmd)) {
			return FTP_ERROR;
		}
		strncpy (session->last_cmd, command, sizeof (session->last_cmd));
	} else if (command[len - 1] == '\r') {
		if (len + 2 > sizeof (session->last_cmd)) {
			return FTP_ERROR;
		}
		strncpy (session->last_cmd, command, sizeof (session->last_cmd));
		session->last_cmd[len] = '\n';
		session->last_cmd[len + 1] = 0;
	} else {
		if (len + 3 > sizeof (session->last_cmd)) {
			return FTP_ERROR;
		}
		strncpy (session->last_cmd, command, sizeof (session->last_cmd));
		session->last_cmd[len] = '\r';
		session->last_cmd[len + 1] = '\n';
		session->last_cmd[len + 2] = 0;
	}
	
	session->estado = FTP_STATE_CMD;
	session->internal = FTP_INTERNAL_WRITE;
	session->command_data_dir = data_dir;
	
	return ftp_session_async_continue (session, FTP_WANT_POLL_OUT);
}

int ftp_session_send_close (FTPSession *session) {
	char *command;
	int len;
	
	if (session->estado != FTP_STATE_LOGGED && session->estado != FTP_STATE_CONNECTED) {
		return FTP_ERROR;
	}
	
	if (session->internal != FTP_INTERNAL_FREE) {
		return FTP_ERROR;
	}
	
	session->estado = FTP_STATE_CLOSING;
	session->internal = FTP_INTERNAL_WRITE;
	
	return ftp_session_async_continue (session, FTP_WANT_POLL_OUT);
}

int ftp_session_request_pasv (FTPSession *session, FILE *stream) {
	char *command;
	int len;
	
	if (session->estado != FTP_STATE_LOGGED) {
		return FTP_ERROR;
	}
	
	if (session->internal != FTP_INTERNAL_FREE) {
		return FTP_ERROR;
	}
	
	session->estado = FTP_STATE_PASV;
	session->internal = FTP_INTERNAL_WRITE;
	session->data_stream = stream;
	
	return ftp_session_async_continue (session, FTP_WANT_POLL_OUT);
}

int ftp_session_get_control_socket (FTPSession *session) {
	return session->csock;
}

int ftp_session_get_data_socket (FTPSession *session) {
	return session->dsock;
}

int ftp_session_get_last_code (FTPSession *session) {
	return session->last_code;
}

char * ftp_session_get_last_message (FTPSession *session) {
	return session->last_message;
}

void ftp_session_set_multiline_func (FTPSession *session, FTPLineFunc func, void *data) {
	session->line_func = func;
	session->line_func_data = data;
}

int ftp_session_get_poll (FTPSession *session) {
	return session->poll;
}

const char *ftp_session_get_user (FTPSession *session) {
	return session->user;
}

